
Input: 
  scene description tmeshes, tmeshesN (one tmesh per object), 
  desired view ppc (a planar pinhole camera)
  framebuffer to hold the image fb

Output:
  fb populated with one color sample per pixel

for each triangle mesh tm
  for each vertex v of tm
    if (programmablePipeline)
      pv = VertexShader(v) // programmable pipeline
    else
      pv = ppc->Project(v) // fixed pipeline

  for each triangle t of tm

    if (programmablePipeline)
      GeometryShader(t)

    aabb = AABB(t.pv0, t.pv1, t.pv2) // compute AABB of projected triangle
    aabb = aabb.Clip(0, 0, fb->w, fb0->h) // clip AABB to screen

    eeqs = SetEdgeEquations(t)
    zABC = SetupZInterpolation(t.v0.z, t.v1.z, t.v2.z)
    r0ABC = SetupMSIN(t.v0.r0, t.v1.r0, t.v2.r0) 
    r1ABC = SetupMSIN(t.v0.r1, t.v1.r1, t.v2.r1)
    ...
    rnABC = SetupMSIN(t.v0.rn, t.v1.rn, t.v2.rn) 
    denABC = SetupMSID() // denominator linear expression for all rasterization
                                  // parameters that are interpolate linearly
                                  // in model space
    for v = aabb.top to aabb.bottom // traverse all rows
      for u = aabb.left to aabb.right // traverse all pixels of current row
        pixv = (u+0.5f, v+0.5f, 1.0f)
        if eeqs[0]*pixv < 0 || eeqs[1]*pixv < 0 || eeqs[2]*pixv < 0
          continue; // outside of triangle
        currz = zABC * pixv
        currR0 = (r0ABC*pixv) / (denABC*pixv)
        currR1 = (r1ABC*pixv) / (denABC*pixv)
        ...
        currRn = (rnABC*pixv) / (denABC*pixv)

        if !programmablePipeline
            currColor = Shade(currR0, currR1, currR2, ..., currRn)
        else
          (currColor, currz) = FragmentShader(currR0, currR1, ..., currRn)
      
        if currz farther than fb->ZBuffer(u, v)
          continue // have already encountered a triangle closer at this pixel
